 {-# OPTIONS -XFlexibleInstances
             -XFlexibleContexts
             -XUndecidableInstances
 #-}

> module Idiom
> where
> import Stdenv
> import qualified Prelude as Stdenv

> class Idiom φ where
>   pure  ::  α -> φ α
>   (◇)   ::  φ (α -> β) -> (φ α -> φ β)
>
>   repeat  ::  α -> φ α
>   map     ::  (α -> β) -> (φ α -> φ β)
>   zip     ::  (α -> β -> γ) -> (φ α -> φ β -> φ γ)
>
>   pure  =  repeat
>   (◇)   =  zip ($)
>
>   repeat a   =  pure a
>   map f s    =  pure f ◇ s
>   zip g s t  =  pure g ◇ s ◇ t

Derived functions.

> (⋆)    ::  (Idiom φ) => φ α -> φ β -> φ (α, β)
> t ⋆ u  =   zip (,) t u

Lifting of functions aka environment monad.

> instance Idiom ((->) k) where
>   pure a  =  \ x -> a
>   f ◇ g   =  \ x -> (f x) (g x)
>
>   repeat a   =  \ x -> a
>   map f h    =  \ x -> f (h x)
>   zip g h i  =  \ x -> g (h x) (i x)

> instance Show (k -> a) where
>   show _  =  "<function>"  -- HACK

> instance Eq (k -> a) where
>   _ == _  =  error "Eq (k -> a): undecidable equality"  -- HACK

> instance (Num a) => Num (k -> a) where
>   (+)          =  zip (+)
>   (-)          =  zip (-)
>   (*)          =  zip (*)
>   negate       =  map negate
>   fromInteger  =  pure . fromInteger
>   abs          =  map abs
>   signum       =  map signum

A generic instance.

> {-
> instance (Idiom f, Num a, Show (f a), Eq (f a)) => Num (f a) where
>   (+)          =  zip (+)
>   (-)          =  zip (-)
>   (*)          =  zip (*)
>   negate       =  map negate
>   fromInteger  =  repeat . fromInteger
>   abs          =  map abs
>   signum       =  map signum
> -}

> unzip    ::  (Idiom f) => f (a, b) -> (f a, f b)
> unzip t  =   (map fst t, map snd t)

