ghci Stream.lhs
ghci -Wall Stream.lhs

> {-# OPTIONS -XMultiParamTypeClasses
>             -XFlexibleInstances
> #-}

> module Stream
> where
> import Stdenv hiding (even, odd, (**))
> import qualified Prelude as Stdenv
> import Memo
> import Idiom
> import NumExt

> default (Integer, Rational)

TODO: does not work in GHC (power nat 2).

%------------------------------------------------------------------------------
\section{Streams}
%------------------------------------------------------------------------------

> data Stream a  =  Cons { head :: a, tail :: Stream a }
 
> infixr 5 ≺
> (≺)    ::  a -> Stream a -> Stream a
> a ≺ s  =   Cons a s

---

Streams of type |Stream α| are memoised functions of type |Natural -> α|.

< Natural -> α  ≅  Stream α 

< data Natural  =  Zero | Succ Natural

< instance Memo Natural Stream where
<   tabulate f  =  f Zero ≺ tabulate (f . Succ)
<
<   lookup s Zero      =  head s
<   lookup s (Succ n)  =  lookup (tail s) n

> instance (Integral a) => Memo a Stream where
>   tabulate f  =  f 0 ≺ tabulate (f . (+ 1))
>
>   lookup s 0        =  head s
>   lookup s (n + 1)  =  lookup (tail s) n
>   lookup _ _        =  error "lookup: negative argument"

---

Stream are idioms aka applicative functors.

> instance Idiom Stream where
>  pure a  =  s where s = a ≺ s
>  s ◇ t   =  (head s) (head t) ≺ (tail s) ◇ (tail t)

>  repeat a   =  s where s = a ≺ s
>  map f s    =  f (head s) ≺ map f (tail s)
>  zip g s t  =  g (head s) (head t) ≺ zip g (tail s) (tail t)

---

Instance declarations.

> len  ::  Int
> len  =   40

> first             ::  Int -> Stream aT -> [aT]
> first 0       _s  =   []
> first (n + 1) s   =   head s : first n (tail s)
> first _       _   =   error "first: negative argument"

> showStream      :: (Show a) => Int -> Stream a -> IO ()
> showStream l s  =  putStrLn (showStreamS l s "")

> showStreamS      :: (Show a) => Int -> Stream a -> ShowS
> showStreamS l s  =  showString "<" . showl (first l s)
>   where showl []        =  showString "..>"
>         showl (a : as)  =  shows a . showString ", " . showl as

> instance (Show a) => Show (Stream a) where
>   showsPrec _d  =  showStreamS len  -- HACK
> instance (Eq a) => Eq (Stream a) where
>   s == t  =  first len s == first len t  -- HACK
> instance (Ord a) => Ord (Stream a) where
>   s <= t  =  first len s <= first len t  -- HACK

`Generic' instances for |Functor| and numeric classes.

> instance Functor Stream where
>   fmap  =  map
> instance (Num a) => Num (Stream a) where
>   (+)          =  zip (+)
>   (-)          =  zip (-)
>   (*)          =  zip (*)
>   negate       =  map negate
>   abs          =  map abs
>   signum       =  map signum
>   fromInteger  =  repeat . fromInteger
> instance (Enum a) => Enum (Stream a) where
>   toEnum i    =  repeat (toEnum i)
>   fromEnum s  =  fromEnum (head s)  -- HACK?!
> instance (Real a) => Real (Stream a) where
>   toRational s  =  toRational (head s)  -- HACK?!
> instance (Integral a) => Integral (Stream a) where
>   div          =  zip div
>   mod          =  zip mod
>   quotRem s t  =  unzip (zip quotRem s t)
>   toInteger s  =  toInteger (head s)  -- HACK?!
> instance (Fractional a) => Fractional (Stream a) where
>   s / t           =  zip (Stdenv./) s t
>   recip s         =  map recip s
>   fromRational r  =  repeat (fromRational r)

> instance (NumExt a) => NumExt (Stream a) where
>   (^)     =  zip (^)
>   (/)     =  zip (/)
>   fact    =  map fact
>   fall    =  zip fall
>   choose  =  zip choose

---

Recursion and iteration.

> recurse      ::  (a -> a) -> (a -> Stream a)
> recurse f a  =   s where s = a ≺ map f s

> iterate      ::  (a -> a) -> (a -> Stream a)
> iterate f a  =   a ≺ iterate f (f a)

%------------------------------------------------------------------------------
\section{Recurrences (|≺|, |⋎|)}
%------------------------------------------------------------------------------

> nat, nat', fac, fib, fib', fib'', fibv, bin, msb, ones, ones', onesv, carry,
>  frac, god, jos :: Stream Integer

NB. The streams can be given the more general type |(Num a) => Stream a|.
However, sometimes this triggers unresolved overloading errors.

> pot, pot' :: Stream Bool

> nat   =  0  ≺ nat + 1
> nat'  =  tail nat
 
> fac   =  1  ≺ (nat + 1) * fac
 
> fib    =  0  ≺ fib'
> fib'   =  1  ≺ fib' + fib
> fib''  =  tail fib'

> fibv  =  0 ≺ fibv + (1 ≺ fibv)

> infixr 5 ⋎
> (⋎)    ::  Stream a -> Stream a -> Stream a
> s ⋎ t  =   head s ≺ t ⋎ tail s

> even, odd  ::  Stream α -> Stream α
> even s     =   head s ≺  odd   (tail s)
> odd  s     =             even  (tail s)

> bin  =  0 ≺ 2 * bin + 1 ⋎ 2 * bin + 2

> pot    =  True ≺ pot ⋎ repeat False
> pot'   =  tail pot
> msb    =  1 ≺ 2 * msb ⋎ 2 * msb
> ones   =  0  ≺ ones'
> ones'  =  1  ≺ ones' ⋎ ones' + 1
> carry  =  0 ⋎ carry + 1

> turn :: (Integral a) => a -> [a]
> turn 0        =  []
> turn (n + 1)  =  turn n ++ [n] ++ turn n
> turn _        =  error "turn: negative argument" 

> tree :: (Integral a) => a -> Stream a
> tree n  =   n ≺ turn n << tree (n + 1)

> infixr 5 <<
> (<<)            ::  [a] -> Stream a -> Stream a
> []        << s  =   s
> (a : as)  << s  =   a ≺ (as << s)

> frac  =  nat ⋎ frac
> god   =  2 * frac + 1
> jos   =  1 ≺ 2 * jos - 1 ⋎ 2 * jos + 1

%-------------------------------=  --------------------------------------------
\section{Finite calculus (|diff|, |sum|)}
%-------------------------------=  --------------------------------------------

> diff    ::  (Num a) => Stream a -> Stream a
> diff s  =   tail s - s

> sum    ::  (Num a) => Stream a -> Stream a
> sum s  =   t where t = 0 ≺ t + s

> onesv  =  0 ≺ onesv + 1 - carry

> sumv    ::  (Num a) => Stream a -> Stream a
> sumv s  =   0 ≺ repeat (head s) + sumv (tail s)

%------------------------------------------------------------------------------
\section{Generating functions (|**|, |//|)}
%------------------------------------------------------------------------------

> const    ::  (Num a) => a -> Stream a
> const n  =   n ≺ repeat 0
 
> z  ::  (Num a) => Stream a
> z  =   0 ≺ 1 ≺ repeat 0

> infixl 7 **
> (**)    ::  (Num a) => Stream a -> Stream a -> Stream a
> s ** t  =   head s * head t ≺ repeat (head s) * tail t + tail s ** t

> reciprocal :: (Fractional a) => Stream a -> Stream a
> reciprocal s  =  t  where  a  =  recip (head s)
>                            t  =  a ≺ repeat (- a) * (tail s ** t)
 
> infixl 7 //
> (//) :: (Fractional a) => Stream a -> Stream a -> Stream a
> s // t  =  s ** reciprocal t

> power :: (Fractional a, Integral b) => Stream a -> b -> Stream a
> power s n
>   | n >= 0     =  pow s n
>   | otherwise  =  reciprocal (pow s (- n))
>   where pow _t 0        =  const 1
>         pow t  (k + 1)  =  t ** pow t k
>         pow _ _         =  error "power: impossible"

%------------------------------------------------------------------------------

Examples.

> main :: IO ()
> main  =  do
>   print $ fib
>   print $ nat * nat
>   print $ tail fib ^ 2 - fib * tail (tail fib)
>   print $ tail fib ^ 2 - fib * tail (tail fib) == (-1) ^ nat
>   print $ pot
>   print $ msb
>   print $ (nat' - msb)
>   print $ ones
>   print $ jos
>   print $ (jos - 1) / 2
>   print $ diff (nat ^ 3)
>   print $ diff (2 ^ nat)
>   print $ carry
>   print $ jos
>   print $ diff (fall nat 3)
>   print $ 3 * fall nat 2
>   print $ sum (0 ⋎ 1 :: Stream Integer) 
>   print $ sum (2 * nat + 1)
>   print $ sum carry
>   print $ nat ** 10 ^ nat
>   print $ 9 * (nat ** 10 ^ nat)
>   print $ 9 * (nat ** 10 ^ nat) + nat'
