module exprSem

/*
	Various executable and testable semantics for a very simple imperative language
	Developpend for CEFP 2009
	Pieter Koopman pieter@cs.ru.nl
*/

import StdEnv, gast

:: AExpr
	= Int Int
	| Var Var
	| (+.) infixl 6 AExpr AExpr
	| (-.) infixl 6 AExpr AExpr
	| (*.) infixl 7 AExpr AExpr
//	| If BExpr AExpr AExpr			// remove comment for exercise 5
//	| (@.) infix Var [AExpr]		// remove comment for exercise 6

:: BExpr
	= TRUE
	| FALSE
	| (=.) infix 4 AExpr AExpr
	| (<.) infix 4 AExpr AExpr
	| ~. BExpr
	| (&&.) infixr 3 BExpr BExpr

:: Stmt
	= (:=.) infix 2 Var AExpr
	| (:.) infixr 1 Stmt Stmt
	| Skip
	| IF BExpr Stmt Stmt
	| While BExpr Stmt
//	| Def Var [Var] AExpr			// remove comment for exercise 6

//:: Res = Val Int | Fun [Var] AExpr	// useful in exercise 6
:: Res   :== Int
:: State :== Var -> Res
:: Var   :== String

A :: AExpr State -> Int
A (Int i)	 s = i
A (Var v)	 s = s v
A (x +. y)	 s = A x s + A y s
A (x -. y)	 s = A x s - A y s
A (x *. y)	 s = A x s * A y s

B :: BExpr State -> Bool
B TRUE      s = True
B FALSE     s = False
B (x =. y)  s = A x s == A y s
B (x <. y)  s = A x s < A y s
B (~. exp)  s = not (B exp s)
B (x &&. y) s = B x s && B y s

(|->) infix :: Var Res -> State -> State
(|->) v i = \env x.if (x==v) i (env x)

emptyState :: State
emptyState = \x -> 0

ns :: Stmt State -> State
ns (v :=. e)   s = (v |-> A e s) s
ns (s1 :. s2)  s = ns s2 (ns s1 s)
ns Skip		   s = s
ns (IF c t e)  s |   B c s  = ns t s
ns (IF c t e)  s | ~(B c s) = ns e s
ns (While c b) s |   B c s  = ns (While c b) (ns b s)
ns (While c b) s | ~(B c s) = s

ds :: Stmt State -> State
ds (v :=. a) s = (v |-> A a s) s
ds Skip      s = s
ds (s1 :. s2) s = ds s2 (ds s1 s)
ds (IF c t e) s = if (B c s) (ds t s) (ds e s)
ds (While c stmt) s = fix f s
where
	f g s = if (B c s) (g (ds stmt s)) s

fix :: (a -> a) -> a
fix f = f (fix f)

:: Config = Final State | Inter Stmt State

sos1 :: Stmt State -> Config
sos1 (v :=. e)  s = Final ((v |-> A e s) s)
sos1 Skip	   s = Final s
sos1 (x :. y) s
 = case sos1 x s of
	Final   t = Inter y t
	Inter z t = Inter (z :. y) t
sos1 (IF c t e) s |   B c s  = Inter t s
sos1 (IF c t e) s | ~(B c s) = Inter e s
sos1 (While c b) s = Inter (IF c (b :. While c b) Skip) s

sosTrace :: Config -> [Config]
sosTrace c=:(Final _) = [c]
sosTrace c=:(Inter ss s) = [c: sosTrace (sos1 ss s)]

sos :: Stmt State -> State
sos s env = env1 where (Final env1) = last (sosTrace (Inter s env)) 

//--------------------

allvars :: Stmt -> [Var]
allvars (x :=. e)  = [x]
allvars (s :. t)   = allvars s + allvars t
allvars Skip       = []
allvars (IF c t e) = allvars t + allvars e
allvars (While c b) = allvars b

//------------------
/*
	no infinite loops (nor very long iterations)
*/
:: Gstmt
	= Assign Gvar GAExpr
	| Comp Gstmt Gstmt
	| Gskip
	| GIF GBExpr Gstmt Gstmt
	| GWhile (Maybe GBExpr) Gvar GNat Gstmt

:: GNat = GNat Int
:: Gvar = Gvar String
:: GAExpr
	= GInt Int
	| GVar Gvar
	| GPlus GAExpr GAExpr
	| GMinus GAExpr GAExpr
	| GTimes GAExpr GAExpr

:: GBExpr
	= GTRUE
	| GFALSE
	| GEqual GAExpr GAExpr
	| GLess GAExpr GAExpr
	| GNot GBExpr
	| GAnd GBExpr GBExpr

derive ggen Gstmt, GAExpr, GBExpr, Maybe
ggen{|GNat|} n r = map GNat [0..]
ggen{|Gvar|} n r = map (Gvar o toString) ['a'..'z']

class conv a b :: a -> b

instance conv GAExpr AExpr
where
	conv (GInt i)		= Int i
	conv (GVar (Gvar v))= Var v
	conv (GPlus x y)	= conv x +. conv y
	conv (GMinus x y)	= conv x -. conv y
	conv (GTimes x y)	= conv x *. conv y
instance conv GBExpr BExpr
where
	conv GTRUE			= TRUE
	conv GFALSE			= FALSE
	conv (GEqual x y)	= conv x =. conv y
	conv (GLess  x y)	= conv x <. conv y
	conv (GNot x)		= ~. (conv x)
	conv (GAnd x y)		= conv x &&. conv y
instance conv Gstmt Stmt
where
	conv (Assign (Gvar v) e)	= v :=. conv e
	conv (Comp x y)				= conv x :. conv y
	conv Gskip					= Skip
	conv (GIF c t e)			= IF (conv c) (conv t) (conv e)
	conv (GWhile b (Gvar v) (GNat n) s)
	 =  counter :=. (Int n) :.
		While cond
	   		(body :.
	   		 counter :=. Var counter -. one
	   		)
	where
		counter = fresh v (allvars body)
		cond = case b of
				Nothing = c0
				Just b	= c0 &&. conv b
		body = conv s
		c0 = Int 0 <. Var counter

fresh :: Var [Var] -> Var
fresh v vs
	| isMember v vs
		= try v vs 0
		= v
where
	try v vs n
		# u = v+toString n
		| isMember u vs
			= try v vs (n+1)
			= u

Start = [v + " = " + toString (state v) + "\n" \\ v <- allvars stmt]
where
	state = ns stmt emptyState
	stmt  = facStmt

facStmt
 =	x :=. Int 4:.
	y :=. one :.
	While (one <. vx)
   	(
   		y :=. vy *. vx :.
   		x :=. vx -. one
   	)
where
	x = "x"; vx = Var x
	y = "y"; vy = Var y

Start1
 =	[quietn 1000 aStream propFacns
	,quietn 1000 aStream propFacds
	,quietn 1000 aStream propFacsos
	,quietn 1000 aStream propFacAll
	,quietn 1000 aStream prop0
	,quietn 1000 aStream (prop1 facStmt)
	,quietn 1000 aStream prop1
	,quietn 1000 aStream prop2ns
	,quietn 1000 aStream prop2ds
	,quietn 1000 aStream prop2sos
	,quietn 1000 aStream prop3
	]
Start2 = test (propFac For [ds,ns,sos])

prop0 :: Stmt -> Property
prop0 stmt = name "prop0" @ eqState (ns stmt emptyState) (ns (Skip :. stmt) emptyState) (allvars stmt)

prop1 :: Stmt -> Property
prop1 stmt = name "prop1" @ eqState (ns stmt emptyState) (ds stmt emptyState) (allvars stmt)

prop2ns :: (Maybe GBExpr) Gvar GNat Gstmt -> Property
prop2ns b v n s = name "prop2ns" @ eqState (ns stmt emptyState) (ns (decl :. IF cond loop Skip) emptyState) (allvars stmt)
where stmt = conv (GWhile b v n s); (decl :. loop=:(While cond body)) = stmt

prop2ds :: (Maybe GBExpr) Gvar GNat Gstmt -> Property
prop2ds b v n s = name "prop2ds" @ eqState (ds stmt emptyState) (ds (decl :. IF cond loop Skip) emptyState) (allvars stmt)
where stmt = conv (GWhile b v n s); (decl :. loop=:(While cond body)) = stmt

prop2sos :: (Maybe GBExpr) Gvar GNat Gstmt -> Property
prop2sos b v n s = name "prop2sos" @ eqState (sos stmt emptyState) (sos (decl :. IF cond loop Skip) emptyState) (allvars stmt)
where stmt = conv (GWhile b v n s); (decl :. loop=:(While cond body)) = stmt

prop3 :: Stmt -> Property
prop3 stmt = name "prop3" @ eqState (ns stmt emptyState) (sos stmt emptyState) (allvars stmt)

propFacns :: Property
propFacns = name "propFac" @ (ns facStmt emptyState "y" == 24)

propFacds :: Property
propFacds = name "propFac" @ (ds facStmt emptyState "y" == 24)

propFacsos :: Property
propFacsos = name "propFac" @ (sos facStmt emptyState "y" == 24)

propFac :: (Stmt State -> State) -> Property
propFac sem = name "propFac" (sem facStmt emptyState "y" == 24)

propFacAll :: Property
propFacAll = name "propFacAll" (propFac For [ns,ds,sos])

eqState :: State State [Var] -> Bool
eqState f g vars = and [f v == g v \\ v<-vars]

instance one AExpr where one = Int 1
instance zero AExpr where zero = Int 0

facState f = emptyState f

loopStmt = While TRUE Skip

ggen{|AExpr|} n r = map conv exprs where exprs :: [GAExpr]; exprs = (ggen{|*|} n r)
ggen{|BExpr|} n r = map conv exprs where exprs :: [GBExpr]; exprs = (ggen{|*|} n r)
ggen{|Stmt|}  n r = map conv stmts where stmts :: [Gstmt] ; stmts = (ggen{|*|} n r)
ggen{|(->)|} _ _ n r = [] // dummy
derive genShow	AExpr, BExpr, Stmt, GNat, Gstmt, GAExpr, GBExpr, Maybe, Gvar
derive gEq		AExpr, BExpr, Stmt
derive bimap	[]

instance + [x] | Eq x where (+) x y = removeDup (x++y)
	